@@ -200,6 +200,10 @@ void proxy_thread_init(LIBEVENT_THREAD *thr) {
     thr->L = L;
     luaL_openlibs(L);
     proxy_register_libs(thr, L);
+    // TODO: srand on time? do we need to bother?
+    for (int x = 0; x < 3; x++) {
+        thr->proxy_rng[x] = rand();
+    }
 
     // kick off the configuration.
     if (proxy_thread_loadconf(thr) != 0) {
@@ -492,7 +496,7 @@ static void _set_noreply_mode(mc_resp *resp, mcp_resp_t *r) {
         case RESP_MODE_METAQUIET:
             if (r->resp.code == MCMC_CODE_MISS) {
                 resp->skip = true;
-            } else if (r->cmd[1] != 'g' && r->resp.code == MCMC_CODE_OK) {
+            } else if (r->cmd != CMD_MG && r->resp.code == MCMC_CODE_OK) {
                 // FIXME (v2): mcmc's parser needs to help us out a bit more
                 // here.
                 // This is a broken case in the protocol though; quiet mode
@@ -523,10 +527,9 @@ int proxy_run_coroutine(lua_State *Lc, mc_resp *resp, io_pending_proxy_t *p, con
 
     if (cores == LUA_OK) {
         WSTAT_DECR(c, proxy_req_active, 1);
-        int type = lua_type(Lc, -1);
+        int type = lua_type(Lc, 1);
         if (type == LUA_TUSERDATA) {
-            mcp_resp_t *r = luaL_checkudata(Lc, -1, "mcp.response");
-            LOGGER_LOG(NULL, LOG_PROXYCMDS, LOGGER_PROXY_RAW, NULL, r->start, r->cmd, r->resp.type, r->resp.code);
+            mcp_resp_t *r = luaL_checkudata(Lc, 1, "mcp.response");
             _set_noreply_mode(resp, r);
             if (r->buf) {
                 // response set from C.
@@ -534,7 +537,7 @@ int proxy_run_coroutine(lua_State *Lc, mc_resp *resp, io_pending_proxy_t *p, con
                 resp->write_and_free = r->buf;
                 resp_add_iov(resp, r->buf, r->blen);
                 r->buf = NULL;
-            } else if (lua_getiuservalue(Lc, -1, 1) != LUA_TNIL) {
+            } else if (lua_getiuservalue(Lc, 1, 1) != LUA_TNIL) {
                 // uservalue slot 1 is pre-created, so we get TNIL instead of
                 // TNONE when nothing was set into it.
                 const char *s = lua_tolstring(Lc, -1, &rlen);
@@ -549,7 +552,7 @@ int proxy_run_coroutine(lua_State *Lc, mc_resp *resp, io_pending_proxy_t *p, con
             }
         } else if (type == LUA_TSTRING) {
             // response is a raw string from lua.
-            const char *s = lua_tolstring(Lc, -1, &rlen);
+            const char *s = lua_tolstring(Lc, 1, &rlen);
             size_t l = rlen > WRITE_BUFFER_SIZE ? WRITE_BUFFER_SIZE : rlen;
             memcpy(resp->wbuf, s, l);
             resp_add_iov(resp, resp->wbuf, l);
@@ -641,8 +644,22 @@ static void proxy_process_command(conn *c, char *command, size_t cmdlen, bool mu
     }
 
     struct proxy_hook *hook = &hooks[pr.command];
+    int hook_ref = hook->lua_ref;
+    // if client came from a tagged listener, scan for a more specific hook.
+    // TODO: (v2) avoiding a hash table lookup here, but maybe some other
+    // datastructure would suffice. for 4-8 tags this is perfectly fast.
+    if (c->tag && hook->tagged) {
+        struct proxy_hook_tagged *pht = hook->tagged;
+        while (pht->lua_ref) {
+            if (c->tag == pht->tag) {
+                hook_ref = pht->lua_ref;
+                break;
+            }
+            pht++;
+        }
+    }
 
-    if (!hook->is_lua) {
+    if (!hook_ref) {
         // need to pass our command string into the internal handler.
         // to minimize the code change, this means allowing it to tokenize the
         // full command. The proxy's indirect parser should be built out to
@@ -751,12 +768,15 @@ static void proxy_process_command(conn *c, char *command, size_t cmdlen, bool mu
     lua_State *Lc = lua_tothread(L, -1);
     // leave the thread first on the stack, so we can reference it if needed.
     // pull the lua hook function onto the stack.
-    lua_rawgeti(Lc, LUA_REGISTRYINDEX, hook->lua_ref);
+    lua_rawgeti(Lc, LUA_REGISTRYINDEX, hook_ref);
 
     mcp_request_t *rq = mcp_new_request(Lc, &pr, command, cmdlen);
     if (multiget) {
         rq->ascii_multiget = true;
     }
+    // NOTE: option 1) copy c->tag into rq->tag here.
+    // add req:listen_tag() to retrieve in top level route.
+
     // TODO (v2): lift this to a post-processor?
     if (rq->pr.vlen != 0) {
         // relying on temporary malloc's not succumbing as poorly to
@@ -802,7 +822,6 @@ static void mcp_queue_io(conn *c, mc_resp *resp, int coro_ref, lua_State *Lc) {
     memset(r, 0, sizeof(mcp_resp_t));
     r->buf = NULL;
     r->blen = 0;
-    r->start = rq->start; // need to inherit the original start time.
     // Set noreply mode.
     // TODO (v2): the response "inherits" the request's noreply mode, which isn't
     // strictly correct; we should inherit based on the request that spawned
@@ -825,15 +844,7 @@ static void mcp_queue_io(conn *c, mc_resp *resp, int coro_ref, lua_State *Lc) {
         r->mode = RESP_MODE_NORMAL;
     }
 
-    int x;
-    int end = rq->pr.reqlen-2 > RESP_CMD_MAX ? RESP_CMD_MAX : rq->pr.reqlen-2;
-    for (x = 0; x < end; x++) {
-        if (rq->pr.request[x] == ' ') {
-            break;
-        }
-        r->cmd[x] = rq->pr.request[x];
-    }
-    r->cmd[x] = '\0';
+    r->cmd = rq->pr.command;
 
     luaL_getmetatable(Lc, "mcp.response");
     lua_setmetatable(Lc, -2);
@@ -868,6 +879,10 @@ static void mcp_queue_io(conn *c, mc_resp *resp, int coro_ref, lua_State *Lc) {
 
     // The direct backend object. Lc is holding the reference in the stack
     p->backend = be;
+    // See #887 for notes.
+    // TODO (v2): hopefully this can be optimized out.
+    strncpy(r->be_name, be->name, MAX_NAMELEN+1);
+    strncpy(r->be_port, be->port, MAX_PORTLEN+1);
 
     mcp_request_attach(Lc, rq, p);
 
