@@ -32,6 +32,45 @@ static int mcplib_response_hit(lua_State *L) {
     return 1;
 }
 
+// Caller needs to discern if a vlen is 0 because of a failed response or an
+// OK response that was actually zero. So we always return an integer value
+// here.
+static int mcplib_response_vlen(lua_State *L) {
+    mcp_resp_t *r = luaL_checkudata(L, -1, "mcp.response");
+
+    // We do remove the "\r\n" from the value length, so if you're actually
+    // processing the value nothing breaks.
+    if (r->resp.vlen >= 2) {
+        lua_pushinteger(L, r->resp.vlen-2);
+    } else {
+        lua_pushinteger(L, 0);
+    }
+
+    return 1;
+}
+
+// Refer to MCMC_CODE_* defines.
+static int mcplib_response_code(lua_State *L) {
+    mcp_resp_t *r = luaL_checkudata(L, -1, "mcp.response");
+
+    lua_pushinteger(L, r->resp.code);
+
+    return 1;
+}
+
+// Get the unparsed response line for handling in lua.
+static int mcplib_response_line(lua_State *L) {
+    mcp_resp_t *r = luaL_checkudata(L, -1, "mcp.response");
+
+    if (r->resp.rline != NULL) {
+        lua_pushlstring(L, r->resp.rline, r->resp.rlen);
+    } else {
+        lua_pushnil(L);
+    }
+
+    return 1;
+}
+
 static int mcplib_response_gc(lua_State *L) {
     mcp_resp_t *r = luaL_checkudata(L, -1, "mcp.response");
 
@@ -80,7 +119,7 @@ static int mcplib_backend(lua_State *L) {
     }
 
     // first check our reference table to compare.
-    lua_pushvalue(L, -4);
+    lua_pushvalue(L, 1);
     int ret = lua_gettable(L, lua_upvalueindex(MCP_BACKEND_UPVALUE));
     if (ret != LUA_TNIL) {
         mcp_backend_t *be_orig = luaL_checkudata(L, -1, "mcp.backend");
@@ -104,7 +143,7 @@ static int mcplib_backend(lua_State *L) {
     strncpy(be->name, name, MAX_NAMELEN);
     strncpy(be->port, port, MAX_PORTLEN);
     be->depth = 0;
-    be->rbuf = NULL;
+    be->rbufused = 0;
     be->failed_count = 0;
     STAILQ_INIT(&be->io_head);
     be->state = mcp_backend_read;
@@ -327,7 +366,9 @@ static int mcplib_pool(lua_State *L) {
         s->ref = luaL_ref(L, LUA_REGISTRYINDEX); // references and pops object.
     }
 
-    if (argc == 1) {
+    // Allow passing an ignored nil as a second argument. Makes the lua easier
+    int type = lua_type(L, 2);
+    if (argc == 1 || type == LUA_TNIL) {
         lua_getglobal(L, "mcp");
         // TODO (v2): decide on a mcp.default_dist and use that instead
         if (lua_getfield(L, -1, "dist_jump_hash") != LUA_TNIL) {
@@ -563,7 +604,7 @@ static int mcplib_attach(lua_State *L) {
     // Pull the original worker thread out of the shared mcplib upvalue.
     LIBEVENT_THREAD *t = lua_touserdata(L, lua_upvalueindex(MCP_THREAD_UPVALUE));
 
-    int hook = luaL_checkinteger(L, -2);
+    int hook = luaL_checkinteger(L, 1);
     // pushvalue to dupe func and etc.
     // can leave original func on stack afterward because it'll get cleared.
     int loop_end = 0;
@@ -579,20 +620,86 @@ static int mcplib_attach(lua_State *L) {
         loop_end = hook + 1;
     }
 
-    if (lua_isfunction(L, -1)) {
+    if (lua_isfunction(L, 2)) {
         struct proxy_hook *hooks = t->proxy_hooks;
+        uint64_t tag = 0; // listener socket tag
+
+        if (lua_isstring(L, 3)) {
+            size_t len;
+            const char *stag = lua_tolstring(L, 3, &len);
+            if (len < 1 || len > 8) {
+                proxy_lua_error(L, "mcp.attach: tag must be 1 to 8 characters");
+                return 0;
+            }
+            memcpy(&tag, stag, len);
+        }
 
         for (int x = loop_start; x < loop_end; x++) {
             struct proxy_hook *h = &hooks[x];
-            lua_pushvalue(L, -1); // duplicate the function for the ref.
-            if (h->lua_ref) {
-                // remove existing reference.
-                luaL_unref(L, LUA_REGISTRYINDEX, h->lua_ref);
-            }
+            lua_pushvalue(L, 2); // duplicate the function for the ref.
 
-            // pops the function from the stack and leaves us a ref. for later.
-            h->lua_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-            h->is_lua = true;
+            if (tag) {
+                // listener was tagged. use the extended hook structure.
+                struct proxy_hook_tagged *pht = h->tagged;
+
+                if (h->tagcount == 0) {
+                    pht = calloc(1, sizeof(struct proxy_hook_tagged));
+                    if (pht == NULL) {
+                        proxy_lua_error(L, "mcp.attach: failure allocating tagged hooks");
+                        return 0;
+                    }
+                    h->tagcount = 1;
+                    h->tagged = pht;
+                }
+
+                bool found = false;
+                for (int x = 0; x < h->tagcount; x++) {
+                    if (pht->tag == tag) {
+                        if (pht->lua_ref) {
+                            // Found existing tagged hook.
+                            luaL_unref(L, LUA_REGISTRYINDEX, pht->lua_ref);
+                        }
+
+                        pht->lua_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+                        assert(pht->lua_ref != 0);
+                        found = true;
+                        break;
+                    } else if (pht->tag == 0) {
+                        // no tag in this slot, so we use it.
+                        pht->lua_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+                        pht->tag = tag;
+                        assert(pht->lua_ref != 0);
+                        found = true;
+                        break;
+                    }
+                    pht++;
+                }
+
+                // need to resize the array to fit the new tag.
+                if (!found) {
+                    pht = realloc(h->tagged, sizeof(struct proxy_hook_tagged) * (h->tagcount+1));
+                    if (!pht) {
+                        proxy_lua_error(L, "mcp.attach: failure to resize tagged hooks");
+                        return 0;
+                    }
+
+                    pht[h->tagcount].lua_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+                    pht[h->tagcount].tag = tag;
+
+                    h->tagcount++;
+                    h->tagged = pht;
+                }
+
+            } else {
+                if (h->lua_ref) {
+                    // remove existing reference.
+                    luaL_unref(L, LUA_REGISTRYINDEX, h->lua_ref);
+                }
+
+                // pops the function from the stack and leaves us a ref. for later.
+                h->lua_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+                assert(h->lua_ref != 0);
+            }
         }
     } else {
         proxy_lua_error(L, "Must pass a function to mcp.attach");
@@ -611,6 +718,125 @@ static int mcplib_log(lua_State *L) {
     return 0;
 }
 
+// (request, resp, "detail")
+static int mcplib_log_req(lua_State *L) {
+    LIBEVENT_THREAD *t = lua_touserdata(L, lua_upvalueindex(MCP_THREAD_UPVALUE));
+    logger *l = t->l;
+    // Not using the LOGGER_LOG macro so we can avoid as much overhead as
+    // possible when logging is disabled.
+    if (! (l->eflags & LOG_PROXYREQS)) {
+        return 0;
+    }
+    int rtype = 0;
+    int rcode = 0;
+    int rstatus = 0;
+    char *rname = NULL;
+    char *rport = NULL;
+
+    mcp_request_t *rq = luaL_checkudata(L, 1, "mcp.request");
+    int type = lua_type(L, 2);
+    if (type == LUA_TUSERDATA) {
+        mcp_resp_t *rs = luaL_checkudata(L, 2, "mcp.response");
+        rtype = rs->resp.type;
+        rcode = rs->resp.code;
+        rstatus = rs->status;
+        rname = rs->be_name;
+        rport = rs->be_port;
+    }
+    size_t dlen = 0;
+    const char *detail = luaL_optlstring(L, 3, NULL, &dlen);
+
+    struct timeval end;
+    gettimeofday(&end, NULL);
+    long elapsed = (end.tv_sec - rq->start.tv_sec) * 1000000 + (end.tv_usec - rq->start.tv_usec);
+
+    logger_log(l, LOGGER_PROXY_REQ, NULL, rq->pr.request, rq->pr.reqlen, elapsed, rtype, rcode, rstatus, detail, dlen, rname, rport);
+
+    return 0;
+}
+
+static inline uint32_t _rotl(const uint32_t x, int k) {
+    return (x << k) | (x >> (32 - k));
+}
+
+// xoroshiro128++ 32bit version.
+static uint32_t _nextrand(uint32_t *s) {
+    const uint32_t result = _rotl(s[0] + s[3], 7) + s[0];
+
+    const uint32_t t = s[1] << 9;
+
+    s[2] ^= s[0];
+    s[3] ^= s[1];
+    s[1] ^= s[2];
+    s[0] ^= s[3];
+
+    s[2] ^= t;
+
+    s[3] = _rotl(s[3], 11);
+
+    return result;
+}
+
+
+// (milliseconds, sample_rate, allerrors, request, resp, "detail")
+static int mcplib_log_reqsample(lua_State *L) {
+    LIBEVENT_THREAD *t = lua_touserdata(L, lua_upvalueindex(MCP_THREAD_UPVALUE));
+    logger *l = t->l;
+    // Not using the LOGGER_LOG macro so we can avoid as much overhead as
+    // possible when logging is disabled.
+    if (! (l->eflags & LOG_PROXYREQS)) {
+        return 0;
+    }
+    int rtype = 0;
+    int rcode = 0;
+    int rstatus = 0;
+    char *rname = NULL;
+    char *rport = NULL;
+
+    int ms = luaL_checkinteger(L, 1);
+    int rate = luaL_checkinteger(L, 2);
+    int allerr = lua_toboolean(L, 3);
+    mcp_request_t *rq = luaL_checkudata(L, 4, "mcp.request");
+    int type = lua_type(L, 5);
+    if (type == LUA_TUSERDATA) {
+        mcp_resp_t *rs = luaL_checkudata(L, 5, "mcp.response");
+        rtype = rs->resp.type;
+        rcode = rs->resp.code;
+        rstatus = rs->status;
+        rname = rs->be_name;
+        rport = rs->be_port;
+    }
+    size_t dlen = 0;
+    const char *detail = luaL_optlstring(L, 6, NULL, &dlen);
+
+    struct timeval end;
+    gettimeofday(&end, NULL);
+    long elapsed = (end.tv_sec - rq->start.tv_sec) * 1000000 + (end.tv_usec - rq->start.tv_usec);
+
+    bool do_log = false;
+    if (allerr && rstatus != MCMC_OK) {
+        do_log = true;
+    } else if (ms > 0 && elapsed > ms * 1000) {
+        do_log = true;
+    } else if (rate > 0) {
+        // slightly biased random-to-rate without adding a loop, which is
+        // completely fine for this use case.
+        uint32_t rnd = (uint64_t)_nextrand(t->proxy_rng) * (uint64_t)rate >> 32;
+        if (rnd == 0) {
+            do_log = true;
+        }
+    }
+
+    if (do_log) {
+        logger_log(l, LOGGER_PROXY_REQ, NULL, rq->pr.request, rq->pr.reqlen, elapsed, rtype, rcode, rstatus, detail, dlen, rname, rport);
+    }
+
+    return 0;
+}
+
+// TODO: slowsample
+// _err versions?
+
 /*** END lua interface to logger ***/
 
 static void proxy_register_defines(lua_State *L) {
@@ -618,6 +844,16 @@ static void proxy_register_defines(lua_State *L) {
     lua_pushinteger(L, x); \
     lua_setfield(L, -2, #x);
 
+    X(MCMC_CODE_STORED);
+    X(MCMC_CODE_EXISTS);
+    X(MCMC_CODE_DELETED);
+    X(MCMC_CODE_TOUCHED);
+    X(MCMC_CODE_VERSION);
+    X(MCMC_CODE_NOT_FOUND);
+    X(MCMC_CODE_NOT_STORED);
+    X(MCMC_CODE_OK);
+    X(MCMC_CODE_NOP);
+    X(MCMC_CODE_MISS);
     X(P_OK);
     X(CMD_ANY);
     X(CMD_ANY_STORAGE);
@@ -646,6 +882,8 @@ int proxy_register_libs(LIBEVENT_THREAD *t, void *ctx) {
         {"rtrimkey", mcplib_request_rtrimkey},
         {"token", mcplib_request_token},
         {"ntokens", mcplib_request_ntokens},
+        {"has_flag", mcplib_request_has_flag},
+        {"flag_token", mcplib_request_flag_token},
         {"__tostring", NULL},
         {"__gc", mcplib_request_gc},
         {NULL, NULL}
@@ -654,6 +892,9 @@ int proxy_register_libs(LIBEVENT_THREAD *t, void *ctx) {
     const struct luaL_Reg mcplib_response_m[] = {
         {"ok", mcplib_response_ok},
         {"hit", mcplib_response_hit},
+        {"vlen", mcplib_response_vlen},
+        {"code", mcplib_response_code},
+        {"line", mcplib_response_line},
         {"__gc", mcplib_response_gc},
         {NULL, NULL}
     };
@@ -678,6 +919,8 @@ int proxy_register_libs(LIBEVENT_THREAD *t, void *ctx) {
         {"stat", mcplib_stat},
         {"await", mcplib_await},
         {"log", mcplib_log},
+        {"log_req", mcplib_log_req},
+        {"log_reqsample", mcplib_log_reqsample},
         {"backend_connect_timeout", mcplib_backend_connect_timeout},
         {"backend_retry_timeout", mcplib_backend_retry_timeout},
         {"backend_read_timeout", mcplib_backend_read_timeout},
