@@ -20,6 +20,7 @@
 
 #include "config.h"
 #include "cache.h"
+#include "crc32c.h"
 #include "hash.h"
 #include "jenkins_hash.h"
 #include "stats_prefix.h"
@@ -96,75 +97,19 @@ static void close_conn() {
 
 static enum test_return cache_create_test(void)
 {
-    cache_t *cache = cache_create("test", sizeof(uint32_t), sizeof(char*),
-                                  NULL, NULL);
+    cache_t *cache = cache_create("test", sizeof(uint32_t), sizeof(char*));
     assert(cache != NULL);
     cache_destroy(cache);
     return TEST_PASS;
 }
 
-const uint64_t constructor_pattern = 0xdeadcafebabebeef;
-
-static int cache_constructor(void *buffer, void *notused1, int notused2) {
-    uint64_t *ptr = buffer;
-    *ptr = constructor_pattern;
-    return 0;
-}
-
-static enum test_return cache_constructor_test(void)
-{
-    cache_t *cache = cache_create("test", sizeof(uint64_t), sizeof(uint64_t),
-                                  cache_constructor, NULL);
-    assert(cache != NULL);
-    uint64_t *ptr = cache_alloc(cache);
-    uint64_t pattern = *ptr;
-    cache_free(cache, ptr);
-    cache_destroy(cache);
-    return (pattern == constructor_pattern) ? TEST_PASS : TEST_FAIL;
-}
-
-static int cache_fail_constructor(void *buffer, void *notused1, int notused2) {
-    return 1;
-}
-
-static enum test_return cache_fail_constructor_test(void)
-{
-    enum test_return ret = TEST_PASS;
-
-    cache_t *cache = cache_create("test", sizeof(uint64_t), sizeof(uint64_t),
-                                  cache_fail_constructor, NULL);
-    assert(cache != NULL);
-    uint64_t *ptr = cache_alloc(cache);
-    if (ptr != NULL) {
-        ret = TEST_FAIL;
-    }
-    cache_destroy(cache);
-    return ret;
-}
-
-static void *destruct_data = 0;
-
-static void cache_destructor(void *buffer, void *notused) {
-    destruct_data = buffer;
-}
-
-static enum test_return cache_destructor_test(void)
-{
-    cache_t *cache = cache_create("test", sizeof(uint32_t), sizeof(char*),
-                                  NULL, cache_destructor);
-    assert(cache != NULL);
-    char *ptr = cache_alloc(cache);
-    cache_free(cache, ptr);
-    cache_destroy(cache);
-
-    return (ptr == destruct_data) ? TEST_PASS : TEST_FAIL;
-}
-
 static enum test_return cache_reuse_test(void)
 {
     int ii;
-    cache_t *cache = cache_create("test", sizeof(uint32_t), sizeof(char*),
-                                  NULL, NULL);
+    cache_t *cache = cache_create("test", sizeof(uint32_t), sizeof(char*));
+    if (cache == NULL) {
+        return TEST_FAIL;
+    }
     char *ptr = cache_alloc(cache);
     cache_free(cache, ptr);
     for (ii = 0; ii < 100; ++ii) {
@@ -179,8 +124,10 @@ static enum test_return cache_reuse_test(void)
 
 static enum test_return cache_bulkalloc(size_t datasize)
 {
-    cache_t *cache = cache_create("test", datasize, sizeof(char*),
-                                  NULL, NULL);
+    cache_t *cache = cache_create("test", datasize, sizeof(char*));
+    if (cache == NULL) {
+        return TEST_FAIL;
+    }
 #define ITERATIONS 1024
     void *ptr[ITERATIONS];
 
@@ -212,9 +159,11 @@ static enum test_return test_issue_161(void)
 static enum test_return cache_redzone_test(void)
 {
 #ifndef HAVE_UMEM_H
-    cache_t *cache = cache_create("test", sizeof(uint32_t), sizeof(char*),
-                                  NULL, NULL);
+    cache_t *cache = cache_create("test", sizeof(uint32_t), sizeof(char*));
 
+    if (cache == NULL) {
+        return TEST_FAIL;
+    }
     /* Ignore SIGABRT */
     struct sigaction old_action;
     struct sigaction action = { .sa_handler = SIG_IGN, .sa_flags = 0};
@@ -249,8 +198,7 @@ static enum test_return cache_limit_revised_downward_test(void)
     int limit = 10, allocated_num = limit + 1, i;
     char ** alloc_objs = calloc(allocated_num, sizeof(char *));
 
-    cache_t *cache = cache_create("test", sizeof(uint32_t), sizeof(char*),
-                                  NULL, NULL);
+    cache_t *cache = cache_create("test", sizeof(uint32_t), sizeof(char*));
     assert(cache != NULL);
 
     /* cache->limit is 0 and we can allocate limit+1 items */
@@ -269,6 +217,8 @@ static enum test_return cache_limit_revised_downward_test(void)
     assert(cache->total == allocated_num-1);
     cache_destroy(cache);
 
+    free(alloc_objs);
+
     return TEST_PASS;
 }
 
@@ -301,6 +251,9 @@ static enum test_return test_stats_prefix_record_get(void) {
 
     stats_prefix_record_get("abc:123", 7, false);
     pfs = stats_prefix_find("abc:123", 7);
+    if (pfs == NULL) {
+        return TEST_FAIL;
+    }
     assert(1 == pfs->num_gets);
     assert(0 == pfs->num_hits);
     stats_prefix_record_get("abc:456", 7, false);
@@ -321,6 +274,9 @@ static enum test_return test_stats_prefix_record_delete(void) {
 
     stats_prefix_record_delete("abc:123", 7);
     pfs = stats_prefix_find("abc:123", 7);
+    if (pfs == NULL) {
+        return TEST_FAIL;
+    }
     assert(0 == pfs->num_gets);
     assert(0 == pfs->num_hits);
     assert(1 == pfs->num_deletes);
@@ -336,6 +292,9 @@ static enum test_return test_stats_prefix_record_set(void) {
 
     stats_prefix_record_set("abc:123", 7);
     pfs = stats_prefix_find("abc:123", 7);
+    if (pfs == NULL) {
+        return TEST_FAIL;
+    }
     assert(0 == pfs->num_gets);
     assert(0 == pfs->num_hits);
     assert(0 == pfs->num_deletes);
@@ -664,7 +623,7 @@ static pid_t start_server(in_port_t *port_out, bool daemon, int timeout) {
 
 static enum test_return test_issue_44(void) {
     in_port_t port;
-    pid_t pid = start_server(&port, true, 15);
+    pid_t pid = start_server(&port, true, 600);
     assert(kill(pid, SIGHUP) == 0);
     sleep(1);
     assert(kill(pid, SIGTERM) == 0);
@@ -880,6 +839,34 @@ static enum test_return test_issue_92(void) {
     return TEST_PASS;
 }
 
+static enum test_return test_crc32c(void) {
+    uint32_t crc_hw, crc_sw;
+
+    char buffer[256];
+    for (int x = 0; x < 256; x++)
+        buffer[x] = x;
+
+    /* Compare hardware to software implementation */
+    crc_hw = crc32c(0, buffer, 256);
+    crc_sw = crc32c_sw(0, buffer, 256);
+    assert(crc_hw == 0x9c44184b);
+    assert(crc_sw == 0x9c44184b);
+
+    /* Test that passing a CRC in also works */
+    crc_hw = crc32c(crc_hw, buffer, 256);
+    crc_sw = crc32c_sw(crc_sw, buffer, 256);
+    assert(crc_hw == 0xae10ee5a);
+    assert(crc_sw == 0xae10ee5a);
+
+    /* Test odd offsets/sizes */
+    crc_hw = crc32c(crc_hw, buffer + 1, 256 - 2);
+    crc_sw = crc32c_sw(crc_sw, buffer + 1, 256 - 2);
+    assert(crc_hw == 0xed37b906);
+    assert(crc_sw == 0xed37b906);
+
+    return TEST_PASS;
+}
+
 static enum test_return test_issue_102(void) {
     char buffer[4096];
     memset(buffer, ' ', sizeof(buffer));
@@ -2026,7 +2013,7 @@ static void *binary_hickup_recv_verification_thread(void *arg) {
 
 static enum test_return test_binary_pipeline_hickup_chunk(void *buffer, size_t buffersize) {
     off_t offset = 0;
-    char *key[256];
+    char *key[256] = { NULL };
     uint64_t value = 0xfeedfacedeadbeef;
 
     while (hickup_thread_running &&
@@ -2260,9 +2247,6 @@ struct testcase {
 
 struct testcase testcases[] = {
     { "cache_create", cache_create_test },
-    { "cache_constructor", cache_constructor_test },
-    { "cache_constructor_fail", cache_fail_constructor_test },
-    { "cache_destructor", cache_destructor_test },
     { "cache_reuse", cache_reuse_test },
     { "cache_redzone", cache_redzone_test },
     { "cache_limit_revised_downward", cache_limit_revised_downward_test },
@@ -2279,6 +2263,7 @@ struct testcase testcases[] = {
     { "issue_44", test_issue_44 },
     { "vperror", test_vperror },
     { "issue_101", test_issue_101 },
+    { "crc32c", test_crc32c },
     /* The following tests all run towards the same server */
     { "start_server", start_memcached_server },
     { "issue_92", test_issue_92 },
@@ -2345,6 +2330,8 @@ int main(int argc, char **argv)
     hash = jenkins_hash;
     stats_prefix_init(':');
 
+    crc32c_init();
+
     for (num_cases = 0; testcases[num_cases].description; num_cases++) {
         /* Just counting */
     }
