@@ -6,6 +6,12 @@
 #ifdef EXTSTORE
 #include "storage.h"
 #endif
+#ifdef HAVE_EVENTFD
+#include <sys/eventfd.h>
+#endif
+#ifdef PROXY
+#include "proto_proxy.h"
+#endif
 #include <assert.h>
 #include <stdio.h>
 #include <errno.h>
@@ -13,6 +19,8 @@
 #include <string.h>
 #include <pthread.h>
 
+#include "queue.h"
+
 #ifdef __sun
 #include <atomic.h>
 #endif
@@ -26,6 +34,14 @@
 /* An item in the connection queue. */
 enum conn_queue_item_modes {
     queue_new_conn,   /* brand new connection. */
+    queue_pause,      /* pause thread */
+    queue_timeout,    /* socket sfd timed out */
+    queue_redispatch, /* return conn from side thread */
+    queue_stop,       /* exit thread */
+    queue_return_io,  /* returning a pending IO object immediately */
+#ifdef PROXY
+    queue_proxy_reload, /* signal proxy to reload worker VM */
+#endif
 };
 typedef struct conn_queue_item CQ_ITEM;
 struct conn_queue_item {
@@ -37,15 +53,18 @@ struct conn_queue_item {
     enum conn_queue_item_modes mode;
     conn *c;
     void    *ssl;
-    CQ_ITEM          *next;
+    uint64_t conntag;
+    enum protocol bproto;
+    io_pending_t *io; // IO when used for deferred IO handling.
+    STAILQ_ENTRY(conn_queue_item) i_next;
 };
 
 /* A connection queue. */
 typedef struct conn_queue CQ;
 struct conn_queue {
-    CQ_ITEM *head;
-    CQ_ITEM *tail;
+    STAILQ_HEAD(conn_ev_head, conn_queue_item) head;
     pthread_mutex_t lock;
+    cache_t *cache; /* freelisted objects */
 };
 
 /* Locks for cache LRU operations */
@@ -64,10 +83,6 @@ static pthread_mutex_t stats_lock = PTHREAD_MUTEX_INITIALIZER;
 /* Lock to cause worker threads to hang up after being woken */
 static pthread_mutex_t worker_hang_lock;
 
-/* Free list of CQ_ITEM structs */
-static CQ_ITEM *cqi_freelist;
-static pthread_mutex_t cqi_freelist_lock;
-
 static pthread_mutex_t *item_locks;
 /* size of the item lock hash table */
 static uint32_t item_lock_count;
@@ -88,6 +103,10 @@ static int init_count = 0;
 static pthread_mutex_t init_lock;
 static pthread_cond_t init_cond;
 
+static void notify_worker(LIBEVENT_THREAD *t, CQ_ITEM *item);
+static void notify_worker_fd(LIBEVENT_THREAD *t, int sfd, enum conn_queue_item_modes mode);
+static CQ_ITEM *cqi_new(CQ *cq);
+static void cq_push(CQ *cq, CQ_ITEM *item);
 
 static void thread_libevent_process(evutil_socket_t fd, short which, void *arg);
 
@@ -137,10 +156,9 @@ static void register_thread_initialized(void) {
 
 /* Must not be called with any deeper locks held */
 void pause_threads(enum pause_thread_types type) {
-    char buf[1];
     int i;
+    bool pause_workers = false;
 
-    buf[0] = 0;
     switch (type) {
         case PAUSE_ALL_THREADS:
             slabs_rebalancer_pause();
@@ -151,7 +169,7 @@ void pause_threads(enum pause_thread_types type) {
             storage_write_pause();
 #endif
         case PAUSE_WORKER_THREADS:
-            buf[0] = 'p';
+            pause_workers = true;
             pthread_mutex_lock(&worker_hang_lock);
             break;
         case RESUME_ALL_THREADS:
@@ -172,17 +190,14 @@ void pause_threads(enum pause_thread_types type) {
     }
 
     /* Only send a message if we have one. */
-    if (buf[0] == 0) {
+    if (!pause_workers) {
         return;
     }
 
     pthread_mutex_lock(&init_lock);
     init_count = 0;
     for (i = 0; i < settings.num_threads; i++) {
-        if (write(threads[i].notify_send_fd, buf, 1) != 1) {
-            perror("Failed writing to notify pipe");
-            /* TODO: This is a fatal problem. Can it ever happen temporarily? */
-        }
+        notify_worker_fd(&threads[i], 0, queue_pause);
     }
     wait_for_thread_registration(settings.num_threads);
     pthread_mutex_unlock(&init_lock);
@@ -193,7 +208,6 @@ void pause_threads(enum pause_thread_types type) {
 // Note: listener thread is the "main" event base, which has exited its
 // loop in order to call this function.
 void stop_threads(void) {
-    char buf[1];
     int i;
 
     // assoc can call pause_threads(), so we have to stop it first.
@@ -203,15 +217,12 @@ void stop_threads(void) {
 
     if (settings.verbose > 0)
         fprintf(stderr, "asking workers to stop\n");
-    buf[0] = 's';
+
     pthread_mutex_lock(&worker_hang_lock);
     pthread_mutex_lock(&init_lock);
     init_count = 0;
     for (i = 0; i < settings.num_threads; i++) {
-        if (write(threads[i].notify_send_fd, buf, 1) != 1) {
-            perror("Failed writing to notify pipe");
-            /* TODO: This is a fatal problem. Can it ever happen temporarily? */
-        }
+        notify_worker_fd(&threads[i], 0, queue_stop);
     }
     wait_for_thread_registration(settings.num_threads);
     pthread_mutex_unlock(&init_lock);
@@ -265,8 +276,12 @@ void stop_threads(void) {
  */
 static void cq_init(CQ *cq) {
     pthread_mutex_init(&cq->lock, NULL);
-    cq->head = NULL;
-    cq->tail = NULL;
+    STAILQ_INIT(&cq->head);
+    cq->cache = cache_create("cq", sizeof(CQ_ITEM), sizeof(char *));
+    if (cq->cache == NULL) {
+        fprintf(stderr, "Failed to create connection queue cache\n");
+        exit(EXIT_FAILURE);
+    }
 }
 
 /*
@@ -278,11 +293,9 @@ static CQ_ITEM *cq_pop(CQ *cq) {
     CQ_ITEM *item;
 
     pthread_mutex_lock(&cq->lock);
-    item = cq->head;
-    if (NULL != item) {
-        cq->head = item->next;
-        if (NULL == cq->head)
-            cq->tail = NULL;
+    item = STAILQ_FIRST(&cq->head);
+    if (item != NULL) {
+        STAILQ_REMOVE_HEAD(&cq->head, i_next);
     }
     pthread_mutex_unlock(&cq->lock);
 
@@ -293,69 +306,72 @@ static CQ_ITEM *cq_pop(CQ *cq) {
  * Adds an item to a connection queue.
  */
 static void cq_push(CQ *cq, CQ_ITEM *item) {
-    item->next = NULL;
-
     pthread_mutex_lock(&cq->lock);
-    if (NULL == cq->tail)
-        cq->head = item;
-    else
-        cq->tail->next = item;
-    cq->tail = item;
+    STAILQ_INSERT_TAIL(&cq->head, item, i_next);
     pthread_mutex_unlock(&cq->lock);
 }
 
 /*
  * Returns a fresh connection queue item.
  */
-static CQ_ITEM *cqi_new(void) {
-    CQ_ITEM *item = NULL;
-    pthread_mutex_lock(&cqi_freelist_lock);
-    if (cqi_freelist) {
-        item = cqi_freelist;
-        cqi_freelist = item->next;
+static CQ_ITEM *cqi_new(CQ *cq) {
+    CQ_ITEM *item = cache_alloc(cq->cache);
+    if (item == NULL) {
+        STATS_LOCK();
+        stats.malloc_fails++;
+        STATS_UNLOCK();
     }
-    pthread_mutex_unlock(&cqi_freelist_lock);
-
-    if (NULL == item) {
-        int i;
-
-        /* Allocate a bunch of items at once to reduce fragmentation */
-        item = malloc(sizeof(CQ_ITEM) * ITEMS_PER_ALLOC);
-        if (NULL == item) {
-            STATS_LOCK();
-            stats.malloc_fails++;
-            STATS_UNLOCK();
-            return NULL;
-        }
-
-        /*
-         * Link together all the new items except the first one
-         * (which we'll return to the caller) for placement on
-         * the freelist.
-         */
-        for (i = 2; i < ITEMS_PER_ALLOC; i++)
-            item[i - 1].next = &item[i];
-
-        pthread_mutex_lock(&cqi_freelist_lock);
-        item[ITEMS_PER_ALLOC - 1].next = cqi_freelist;
-        cqi_freelist = &item[1];
-        pthread_mutex_unlock(&cqi_freelist_lock);
-    }
-
     return item;
 }
 
-
 /*
  * Frees a connection queue item (adds it to the freelist.)
  */
-static void cqi_free(CQ_ITEM *item) {
-    pthread_mutex_lock(&cqi_freelist_lock);
-    item->next = cqi_freelist;
-    cqi_freelist = item;
-    pthread_mutex_unlock(&cqi_freelist_lock);
+static void cqi_free(CQ *cq, CQ_ITEM *item) {
+    cache_free(cq->cache, item);
 }
 
+// TODO: Skip notify if queue wasn't empty?
+// - Requires cq_push() returning a "was empty" flag
+// - Requires event handling loop to pop the entire queue and work from that
+// instead of the ev_count work there now.
+// In testing this does result in a large performance uptick, but unclear how
+// much that will transfer from a synthetic benchmark.
+static void notify_worker(LIBEVENT_THREAD *t, CQ_ITEM *item) {
+    cq_push(t->ev_queue, item);
+#ifdef HAVE_EVENTFD
+    uint64_t u = 1;
+    if (write(t->notify_event_fd, &u, sizeof(uint64_t)) != sizeof(uint64_t)) {
+        perror("failed writing to worker eventfd");
+        /* TODO: This is a fatal problem. Can it ever happen temporarily? */
+    }
+#else
+    char buf[1] = "c";
+    if (write(t->notify_send_fd, buf, 1) != 1) {
+        perror("Failed writing to notify pipe");
+        /* TODO: This is a fatal problem. Can it ever happen temporarily? */
+    }
+#endif
+}
+
+// NOTE: An external func that takes a conn *c might be cleaner overall.
+static void notify_worker_fd(LIBEVENT_THREAD *t, int sfd, enum conn_queue_item_modes mode) {
+    CQ_ITEM *item;
+    while ( (item = cqi_new(t->ev_queue)) == NULL ) {
+        // NOTE: most callers of this function cannot fail, but mallocs in
+        // theory can fail. Small mallocs essentially never do without also
+        // killing the process. Syscalls can also fail but the original code
+        // never handled this either.
+        // As a compromise, I'm leaving this note and this loop: This alloc
+        // cannot fail, but pre-allocating the data is too much code in an
+        // area I want to keep more lean. If this CQ business becomes a more
+        // generic queue I'll reconsider.
+    }
+
+    item->mode = mode;
+    item->sfd = sfd;
+    notify_worker(t, item);
+}
 
 /*
  * Creates a worker thread.
@@ -403,8 +419,13 @@ static void setup_thread(LIBEVENT_THREAD *me) {
     }
 
     /* Listen for notifications from other threads */
+#ifdef HAVE_EVENTFD
+    event_set(&me->notify_event, me->notify_event_fd,
+              EV_READ | EV_PERSIST, thread_libevent_process, me);
+#else
     event_set(&me->notify_event, me->notify_receive_fd,
               EV_READ | EV_PERSIST, thread_libevent_process, me);
+#endif
     event_base_set(me->base, &me->notify_event);
 
     if (event_add(&me->notify_event, 0) == -1) {
@@ -412,19 +433,19 @@ static void setup_thread(LIBEVENT_THREAD *me) {
         exit(1);
     }
 
-    me->new_conn_queue = malloc(sizeof(struct conn_queue));
-    if (me->new_conn_queue == NULL) {
+    me->ev_queue = malloc(sizeof(struct conn_queue));
+    if (me->ev_queue == NULL) {
         perror("Failed to allocate memory for connection queue");
         exit(EXIT_FAILURE);
     }
-    cq_init(me->new_conn_queue);
+    cq_init(me->ev_queue);
 
     if (pthread_mutex_init(&me->stats.mutex, NULL) != 0) {
         perror("Failed to initialize mutex");
         exit(EXIT_FAILURE);
     }
 
-    me->rbuf_cache = cache_create("rbuf", READ_BUFFER_SIZE, sizeof(char *), NULL, NULL);
+    me->rbuf_cache = cache_create("rbuf", READ_BUFFER_SIZE, sizeof(char *));
     if (me->rbuf_cache == NULL) {
         fprintf(stderr, "Failed to create read buffer cache\n");
         exit(EXIT_FAILURE);
@@ -441,13 +462,11 @@ static void setup_thread(LIBEVENT_THREAD *me) {
         cache_set_limit(me->rbuf_cache, limit);
     }
 
-#ifdef EXTSTORE
-    me->io_cache = cache_create("io", sizeof(io_wrap), sizeof(char*), NULL, NULL);
+    me->io_cache = cache_create("io", sizeof(io_pending_t), sizeof(char*));
     if (me->io_cache == NULL) {
         fprintf(stderr, "Failed to create IO object cache\n");
         exit(EXIT_FAILURE);
     }
-#endif
 #ifdef TLS
     if (settings.ssl_enabled) {
         me->ssl_wbuf = (char *)malloc((size_t)settings.ssl_wbuf_size);
@@ -457,6 +476,24 @@ static void setup_thread(LIBEVENT_THREAD *me) {
         }
     }
 #endif
+#ifdef EXTSTORE
+    // me->storage is set just before this function is called.
+    if (me->storage) {
+        thread_io_queue_add(me, IO_QUEUE_EXTSTORE, me->storage,
+            storage_submit_cb, storage_complete_cb, NULL, storage_finalize_cb);
+    }
+#endif
+#ifdef PROXY
+    thread_io_queue_add(me, IO_QUEUE_PROXY, settings.proxy_ctx, proxy_submit_cb,
+            proxy_complete_cb, proxy_return_cb, proxy_finalize_cb);
+
+    // TODO: maybe register hooks to be called here from sub-packages? ie;
+    // extstore, TLS, proxy.
+    if (settings.proxy_enabled) {
+        proxy_thread_init(me);
+    }
+#endif
+    thread_io_queue_add(me, IO_QUEUE_NONE, NULL, NULL, NULL, NULL, NULL);
 }
 
 /*
@@ -491,34 +528,48 @@ static void *worker_libevent(void *arg) {
 
 
 /*
- * Processes an incoming "handle a new connection" item. This is called when
+ * Processes an incoming "connection event" item. This is called when
  * input arrives on the libevent wakeup pipe.
  */
+// Syscalls can be expensive enough that handling a few of them once here can
+// save both throughput and overall latency.
+#define MAX_PIPE_EVENTS 32
 static void thread_libevent_process(evutil_socket_t fd, short which, void *arg) {
     LIBEVENT_THREAD *me = arg;
     CQ_ITEM *item;
-    char buf[1];
     conn *c;
-    unsigned int fd_from_pipe;
-
-    if (read(fd, buf, 1) != 1) {
+    uint64_t ev_count = 0; // max number of events to loop through this run.
+#ifdef HAVE_EVENTFD
+    // NOTE: unlike pipe we aren't limiting the number of events per read.
+    // However we do limit the number of queue pulls to what the count was at
+    // the time of this function firing.
+    if (read(fd, &ev_count, sizeof(uint64_t)) != sizeof(uint64_t)) {
         if (settings.verbose > 0)
             fprintf(stderr, "Can't read from libevent pipe\n");
         return;
     }
+#else
+    char buf[MAX_PIPE_EVENTS];
 
-    switch (buf[0]) {
-    case 'c':
-        item = cq_pop(me->new_conn_queue);
+    ev_count = read(fd, buf, MAX_PIPE_EVENTS);
+    if (ev_count == 0) {
+        if (settings.verbose > 0)
+            fprintf(stderr, "Can't read from libevent pipe\n");
+        return;
+    }
+#endif
 
-        if (NULL == item) {
-            break;
+    for (int x = 0; x < ev_count; x++) {
+        item = cq_pop(me->ev_queue);
+        if (item == NULL) {
+            return;
         }
+
         switch (item->mode) {
             case queue_new_conn:
                 c = conn_new(item->sfd, item->init_state, item->event_flags,
                                    item->read_buffer_size, item->transport,
-                                   me->base, item->ssl);
+                                   me->base, item->ssl, item->conntag, item->bproto);
                 if (c == NULL) {
                     if (IS_UDP(item->transport)) {
                         fprintf(stderr, "Can't listen for events on UDP socket\n");
@@ -538,6 +589,7 @@ static void thread_libevent_process(evutil_socket_t fd, short which, void *arg)
                     }
                 } else {
                     c->thread = me;
+                    conn_io_queue_setup(c);
 #ifdef TLS
                     if (settings.ssl_enabled && c->ssl != NULL) {
                         assert(c->thread && c->thread->ssl_wbuf);
@@ -546,50 +598,134 @@ static void thread_libevent_process(evutil_socket_t fd, short which, void *arg)
 #endif
                 }
                 break;
+            case queue_pause:
+                /* we were told to pause and report in */
+                register_thread_initialized();
+                break;
+            case queue_timeout:
+                /* a client socket timed out */
+                conn_close_idle(conns[item->sfd]);
+                break;
+            case queue_redispatch:
+                /* a side thread redispatched a client connection */
+                conn_worker_readd(conns[item->sfd]);
+                break;
+            case queue_stop:
+                /* asked to stop */
+                event_base_loopexit(me->base, NULL);
+                break;
+            case queue_return_io:
+                /* getting an individual IO object back */
+                conn_io_queue_return(item->io);
+                break;
+#ifdef PROXY
+            case queue_proxy_reload:
+                proxy_worker_reload(settings.proxy_ctx, me);
+                break;
+#endif
         }
-        cqi_free(item);
-        break;
-    /* we were told to pause and report in */
-    case 'p':
-        register_thread_initialized();
-        break;
-    /* a client socket timed out */
-    case 't':
-        if (read(fd, &fd_from_pipe, sizeof(fd_from_pipe)) != sizeof(fd_from_pipe)) {
-            if (settings.verbose > 0)
-                fprintf(stderr, "Can't read timeout fd from libevent pipe\n");
-            return;
-        }
-        conn_close_idle(conns[fd_from_pipe]);
-        break;
-    /* a side thread redispatched a client connection */
-    case 'r':
-        if (read(fd, &fd_from_pipe, sizeof(fd_from_pipe)) != sizeof(fd_from_pipe)) {
-            if (settings.verbose > 0)
-                fprintf(stderr, "Can't read redispatch fd from libevent pipe\n");
-            return;
-        }
-        conn_worker_readd(conns[fd_from_pipe]);
-        break;
-    /* asked to stop */
-    case 's':
-        event_base_loopexit(me->base, NULL);
-        break;
+
+        cqi_free(me->ev_queue, item);
     }
 }
 
+// NOTE: need better encapsulation.
+// used by the proxy module to iterate the worker threads.
+LIBEVENT_THREAD *get_worker_thread(int id) {
+    return &threads[id];
+}
+
 /* Which thread we assigned a connection to most recently. */
 static int last_thread = -1;
 
+/* Last thread we assigned to a connection based on napi_id */
+static int last_thread_by_napi_id = -1;
+
+static LIBEVENT_THREAD *select_thread_round_robin(void)
+{
+    int tid = (last_thread + 1) % settings.num_threads;
+
+    last_thread = tid;
+
+    return threads + tid;
+}
+
+static void reset_threads_napi_id(void)
+{
+    LIBEVENT_THREAD *thread;
+    int i;
+
+    for (i = 0; i < settings.num_threads; i++) {
+         thread = threads + i;
+         thread->napi_id = 0;
+    }
+
+    last_thread_by_napi_id = -1;
+}
+
+/* Select a worker thread based on the NAPI ID of an incoming connection
+ * request. NAPI ID is a globally unique ID that identifies a NIC RX queue
+ * on which a flow is received.
+ */
+static LIBEVENT_THREAD *select_thread_by_napi_id(int sfd)
+{
+    LIBEVENT_THREAD *thread;
+    int napi_id, err, i;
+    socklen_t len;
+    int tid = -1;
+
+    len = sizeof(socklen_t);
+    err = getsockopt(sfd, SOL_SOCKET, SO_INCOMING_NAPI_ID, &napi_id, &len);
+    if ((err == -1) || (napi_id == 0)) {
+        STATS_LOCK();
+        stats.round_robin_fallback++;
+        STATS_UNLOCK();
+        return select_thread_round_robin();
+    }
+
+select:
+    for (i = 0; i < settings.num_threads; i++) {
+         thread = threads + i;
+         if (last_thread_by_napi_id < i) {
+             thread->napi_id = napi_id;
+             last_thread_by_napi_id = i;
+             tid = i;
+             break;
+         }
+         if (thread->napi_id == napi_id) {
+             tid = i;
+             break;
+         }
+    }
+
+    if (tid == -1) {
+        STATS_LOCK();
+        stats.unexpected_napi_ids++;
+        STATS_UNLOCK();
+        reset_threads_napi_id();
+        goto select;
+    }
+
+    return threads + tid;
+}
+
 /*
  * Dispatches a new connection to another thread. This is only ever called
  * from the main thread, either during initialization (for UDP) or because
  * of an incoming connection.
  */
 void dispatch_conn_new(int sfd, enum conn_states init_state, int event_flags,
-                       int read_buffer_size, enum network_transport transport, void *ssl) {
-    CQ_ITEM *item = cqi_new();
-    char buf[1];
+                       int read_buffer_size, enum network_transport transport, void *ssl,
+                       uint64_t conntag, enum protocol bproto) {
+    CQ_ITEM *item = NULL;
+    LIBEVENT_THREAD *thread;
+
+    if (!settings.num_napi_ids)
+        thread = select_thread_round_robin();
+    else
+        thread = select_thread_by_napi_id(sfd);
+
+    item = cqi_new(thread->ev_queue);
     if (item == NULL) {
         close(sfd);
         /* given that malloc failed this may also fail, but let's try */
@@ -597,12 +733,6 @@ void dispatch_conn_new(int sfd, enum conn_states init_state, int event_flags,
         return;
     }
 
-    int tid = (last_thread + 1) % settings.num_threads;
-
-    LIBEVENT_THREAD *thread = threads + tid;
-
-    last_thread = tid;
-
     item->sfd = sfd;
     item->init_state = init_state;
     item->event_flags = event_flags;
@@ -610,30 +740,44 @@ void dispatch_conn_new(int sfd, enum conn_states init_state, int event_flags,
     item->transport = transport;
     item->mode = queue_new_conn;
     item->ssl = ssl;
-
-    cq_push(thread->new_conn_queue, item);
+    item->conntag = conntag;
+    item->bproto = bproto;
 
     MEMCACHED_CONN_DISPATCH(sfd, (int64_t)thread->thread_id);
-    buf[0] = 'c';
-    if (write(thread->notify_send_fd, buf, 1) != 1) {
-        perror("Writing to thread notify pipe");
-    }
+    notify_worker(thread, item);
 }
 
 /*
  * Re-dispatches a connection back to the original thread. Can be called from
  * any side thread borrowing a connection.
  */
-#define REDISPATCH_MSG_SIZE (1 + sizeof(int))
 void redispatch_conn(conn *c) {
-    char buf[REDISPATCH_MSG_SIZE];
-    LIBEVENT_THREAD *thread = c->thread;
+    notify_worker_fd(c->thread, c->sfd, queue_redispatch);
+}
 
-    buf[0] = 'r';
-    memcpy(&buf[1], &c->sfd, sizeof(int));
-    if (write(thread->notify_send_fd, buf, REDISPATCH_MSG_SIZE) != REDISPATCH_MSG_SIZE) {
-        perror("Writing redispatch to thread notify pipe");
+void timeout_conn(conn *c) {
+    notify_worker_fd(c->thread, c->sfd, queue_timeout);
+}
+#ifdef PROXY
+void proxy_reload_notify(LIBEVENT_THREAD *t) {
+    notify_worker_fd(t, 0, queue_proxy_reload);
+}
+#endif
+
+void return_io_pending(io_pending_t *io) {
+    CQ_ITEM *item = cqi_new(io->thread->ev_queue);
+    if (item == NULL) {
+        // TODO: how can we avoid this?
+        // In the main case I just loop, since a malloc failure here for a
+        // tiny object that's generally in a fixed size queue is going to
+        // implode shortly.
+        return;
     }
+
+    item->mode = queue_return_io;
+    item->io = io;
+
+    notify_worker(io->thread, item);
 }
 
 /* This misses the allow_new_conns flag :( */
@@ -791,6 +935,9 @@ void threadlocal_stats_reset(void) {
 #ifdef EXTSTORE
         EXTSTORE_THREAD_STATS_FIELDS
 #endif
+#ifdef PROXY
+        PROXY_THREAD_STATS_FIELDS
+#endif
 #undef X
 
         memset(&threads[ii].stats.slab_stats, 0,
@@ -816,6 +963,9 @@ void threadlocal_stats_aggregate(struct thread_stats *stats) {
 #ifdef EXTSTORE
         EXTSTORE_THREAD_STATS_FIELDS
 #endif
+#ifdef PROXY
+        PROXY_THREAD_STATS_FIELDS
+#endif
 #undef X
 
         for (sid = 0; sid < MAX_NUMBER_OF_SLAB_CLASSES; sid++) {
@@ -868,9 +1018,6 @@ void memcached_thread_init(int nthreads, void *arg) {
     pthread_mutex_init(&init_lock, NULL);
     pthread_cond_init(&init_cond, NULL);
 
-    pthread_mutex_init(&cqi_freelist_lock, NULL);
-    cqi_freelist = NULL;
-
     /* Want a wide lock table, but don't waste memory */
     if (nthreads < 3) {
         power = 10;
@@ -913,6 +1060,13 @@ void memcached_thread_init(int nthreads, void *arg) {
     }
 
     for (i = 0; i < nthreads; i++) {
+#ifdef HAVE_EVENTFD
+        threads[i].notify_event_fd = eventfd(0, EFD_NONBLOCK);
+        if (threads[i].notify_event_fd == -1) {
+            perror("failed creating eventfd for worker thread");
+            exit(1);
+        }
+#else
         int fds[2];
         if (pipe(fds)) {
             perror("Can't create notify pipe");
@@ -921,6 +1075,7 @@ void memcached_thread_init(int nthreads, void *arg) {
 
         threads[i].notify_receive_fd = fds[0];
         threads[i].notify_send_fd = fds[1];
+#endif
 #ifdef EXTSTORE
         threads[i].storage = arg;
 #endif
